# This module handles CRUD for the sqlite database

# sqlalchemy
import sqlalchemy as db
from sqlalchemy import MetaData
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy import Integer, String

from sqlalchemy.orm import Session, sessionmaker
from sqlalchemy import inspect, select, update, delete, values

# marshmallow 
from marshmallow_sqlalchemy import SQLAlchemyAutoSchema

# Custom modules
import ao3scraper.constants as constants

engine = db.create_engine(f'sqlite:///{constants.DATABASE_FILE_PATH}')
connection = engine.connect()

Base = declarative_base()
metadata = MetaData()

# Setup sessionmaker so we don't need to use Session(engine) every time
Session = sessionmaker(engine)

# Create fanfic class
class Fanfic(Base):
    __tablename__ = "fics"
    
    id = db.Column(Integer, primary_key=True, unique=True)

# Add each item in TABLE_COLUMNS to the database as a db.Column(String) object.
for column in constants.TABLE_COLUMNS:
    setattr(Fanfic, column, db.Column(String))

# Create Schema for serialization via Marshmallow.
class FanficSchema(SQLAlchemyAutoSchema):
    class Meta:
        model = Fanfic
        include_relationships = True
        load_instance = True


def add_fic(fic_id):
    """Adds a new fic to the database."""
    with Session() as session:
        session.add(Fanfic(id=fic_id))
        session.commit()

def add_all_fics(fic_id):
    """Adds multiple new fics to the database. (Input needs to be list)"""

    # https://stackoverflow.com/questions/7889183/sqlalchemy-insert-or-update-example
    with Session() as session:
        session.add_all(Fanfic(id=fic_id))
        session.commit()


def update_fic(fic, fic_id):
    """Updates all fics with scraped data."""

    with Session() as session:
        stmt = update(Fanfic).where(Fanfic.id == int(fic_id)).values(fic)
        session.execute(stmt)
        session.commit()


def delete_fic(fic_id):
    """Deletes a fic from the database."""

    with Session() as session:
        query = session.get(Fanfic, fic_id)
        session.delete(query)
        session.commit()

def get_fic(fic_id):
    """Returns a fic from the database."""

    with Session() as session:
        query = session.get(Fanfic, fic_id)
        result_dict = FanficSchema().dump(query)
        return result_dict


def get_all_fics():
    """Returns all fics from the database."""

    with Session() as session:
        query = session.query(Fanfic)
        result_dict = [FanficSchema().dump(u) for u in query.all()]
        return result_dict
    

def get_fic_ids():
    """Returns all fic IDs from the database in a list."""

    with Session() as session:
        query = session.query(Fanfic.id).all()
        result = [r for r, in query]
        return result




# This module validates that MySQL database (fics.db) and the config file (config.yaml) exist, if not, it creates them.

import sqlite3
from os import path
import pathlib
#from yaml import dump, Dumper
from ruamel.yaml import YAML

from sqlite3 import OperationalError
from alembic.config import Config
from alembic import command
import uuid
from shutil import copy

# Import custom modules
import ao3scraper.constants as constants

def main():
    # Create config file if config file does not exist
    if not path.exists(constants.CONFIG_FILE_PATH):
        print("No config file found. Creating new config file...")

        pathlib.Path(constants.CONFIG_PATH).mkdir(parents=True, exist_ok=True)
        
        with open(constants.CONFIG_FILE_PATH, 'w') as file:
            # Convert CONFIG_TEMPLATE to yaml, and disable the alphabetical key sorting done by yaml.dump
            #config_file_dump = dump(constants.CONFIG_TEMPLATE, Dumper=Dumper, sort_keys=False)
            yaml = YAML()
            template = yaml.load(constants.CONFIG_TEMPLATE)
            # Write to file
            yaml.dump(template, file)
            #file.write(yaml.dump(template))
            pass

        print("Config file created.")
        print("You can change configuration options in config.yaml")

    # Create database file if database does not exist
    if not path.exists(constants.DATABASE_FILE_PATH):
        print("No database found. Creating new database...")

        pathlib.Path(constants.DATA_PATH).mkdir(parents=True, exist_ok=True)
 
        # Connect to database
        connection = sqlite3.connect(constants.DATABASE_FILE_PATH)
        cursor = connection.cursor()

        # Create fics table
        cursor.execute("CREATE TABLE fics (id INTEGER)")
        for column in constants.TABLE_COLUMNS:
            cursor.execute(f"ALTER TABLE fics ADD {column} TEXT")
            
        # Create metadata table
        cursor.execute("CREATE TABLE alembic_version (version_num VARCHAR(32) NOT NULL, CONSTRAINT alembic_version_pkc PRIMARY KEY (version_num));")
        query = f"INSERT INTO alembic_version VALUES ('{constants.ALEMBIC_VERSION}');"
        cursor.execute(query)

        connection.commit()
        connection.close()

        print("Database created.\n")
    
    validate_database()

def validate_database():
    # A new database has NOT been created, so we must validate the existing one.
    # Connect to database
    connection = sqlite3.connect(constants.DATABASE_FILE_PATH)
    cursor = connection.cursor()

    # Retrive version_num
    try:
        cursor.execute("SELECT version_num FROM alembic_version;")
        version_num = [item[0] for item in cursor.fetchall()]
        version_num = version_num[0]
    except OperationalError:
        print(f"The local database version could not be verified. It may be a legacy version.")
        upgrade_database()
    except IndexError:
        print("The local database version entry does not exist. This may be because of a previous database migration that failed.")
        exit()

    if version_num != constants.ALEMBIC_VERSION:
        print(f"ao3scraper is on database revision {constants.ALEMBIC_VERSION}, but the local database is on revision {version_num}.")
        upgrade_database()
    
def upgrade_database():
    print("Would you like to migrate to the version supported by ao3scraper? (y/n)")
    
    choice = input(" > ").lower()
    if choice == 'y':
        backup_db_name = "fics_backup_" + str(uuid.uuid4().hex) + ".db"
        backup_db_path = path.join(constants.DATA_PATH, 'backups', backup_db_name)

        pathlib.Path(path.join(constants.DATA_PATH, 'backups')).mkdir(parents=False, exist_ok=True)
        print(f"Saving backup of current database to {backup_db_path}")
        copy(constants.DATABASE_FILE_PATH, backup_db_path)

        alembic_cfg = Config("alembic.ini")
        command.upgrade(alembic_cfg, constants.ALEMBIC_VERSION)
        quit()
    else:
        # Quitting is the safest option, as ao3scraper doesn't do error handling for the database very well.
        quit()


import AO3

"""
vars(work) -> return every attribute of object, along with values
AO3.Work.metadata -> returns the property object
AO3.Work(X).metadata

inspect.getmembers(AO3.Work.metadata)

AO3.common.get_work_from_banner(workid)

for item in attributes:
    print(type(attributes[i]))

list(work.metadata) -> get keys as list
work.metadata.keys() -> get keys as dict_keys
"""

# Retrive example AO3 work (This can be any, but a smaller one is quicker.)
work = AO3.Work(26754208, load_chapters=False)

# List each key in work.metadata
key_list = list(work.metadata.keys())

TABLE_COLUMNS = []

# Exclude 'id' from list
for item in key_list:
    if item != 'id':
        TABLE_COLUMNS.append(item)

# Print final columns
print(TABLE_COLUMNS)


# The main program for ao3scraper.

# Core modules
import AO3
import click

# Web scraping
import requests
from requests.exceptions import Timeout, ConnectionError
from concurrent.futures import ThreadPoolExecutor

# Formatting
from rich.console import Console
from rich.table import Table
from rich.progress import Progress
from rich.traceback import install
install()

# Other modules
from datetime import datetime
import copy
import pickle
import warnings

# Custom modules
import ao3scraper.constants as constants
import ao3scraper.database as database

# Create columns of rich table
table = Table(title="Fanfics", show_lines=True)
table.add_column("Index")

args = []
for i in constants.TABLE_TEMPLATE:
    table.add_column(i["name"], style=i["styles"])
    args.append(i["column"])

console = Console()

# Read database
fic_ids = database.get_fic_ids()
local_fics = database.get_all_fics()

# Start click command
@click.command()
@click.option('--scrape', '-s', is_flag=True, help='Launches scraping mode.')
@click.option('--cache', '-c', is_flag=True, help='Prints the last scraped table.')
@click.option('--list', '-l', is_flag=True, help='Lists all entries in the database.')
@click.option('--add', '-a', help='Adds a single url to the database.')
@click.option('--add-urls', is_flag=True, help='Opens a text file to add multiple urls to the database.')
@click.option('--delete', '-d', help='Deletes an entry from the database.', type=int)
@click.option('--version', '-v', is_flag=True, help='Display version of ao3scraper and other info.')
def main(scrape, cache, list, add, add_urls, delete, version):
    if scrape:
        scrape_urls()
    elif cache:
        print_cached_table()
    elif list:
        construct_rich_table(list)
    elif add:
        add_url_single(add)
    elif add_urls:
        add_url_multiple()
    elif delete:
        delete_entry(delete)
    elif version:
        print(
            f"Version: {constants.APP_VERSION}\n"
            f"Data location: {constants.DATA_PATH}\n"
            f"Config location: {constants.CONFIG_PATH}\n"
        )
        console.print("Made with :heart: by [link=https://github.com/EthanLeitch]Ethan Leitch[/link].")
    else:
        with click.Context(main, info_name='ao3scraper.py') as ctx:
            click.echo(main.get_help(ctx))


def scrape_urls():
    # Check if AO3 is online / accessible
    print("Checking if AO3 servers are online...")
    try:
        requests.get("https://archiveofourown.org", timeout=10)
    except Timeout:
        print("Could not reach AO3 servers. (Timeout)")
        exit()
    except ConnectionError:
        print("Could not reach AO3 servers. (ConnectionError)")
        exit()
    else:
        print("Contacted servers successfully.")
    
    # Create array of None that will be replaced later depending on the thread's index.
    external_fics = [None for fic in fic_ids]

    # The load_fic function that each thread performs
    def load_fic(id):
        try:
            # Setup custom warning format
            def custom_formatwarning(msg, *args, **kwargs):
                if constants.WARNINGS:
                    return f"(Work: {id}) Warning: {str(msg)}\n"
                return ""

            warnings.formatwarning = custom_formatwarning

            # Fetch fic's metadata
            fic = AO3.Work(id).metadata

            # Place each fic in correct array position. NOTE: This may not be particularly efficient as the entire list is enumerated by each thread.
            for count, item in enumerate(fic_ids):
                if fic["id"] == item:
                    external_fics[count] = fic

        except Exception as e:
            # print(f"{e} {id}")
            for count, local_id in enumerate(fic_ids):
                if id == local_id:
                    if type(e) is AttributeError:
                        e = "The work might be restricted (AttributeError)"

                    external_fics[count] = ({'Exception': str(e), 'id': id})
            
            progress.update(progress_bar, advance=1)
        
        progress.update(progress_bar, advance=1)
    
    # Track and create thread pool
    with Progress() as progress:
        progress_bar = progress.add_task("Fetching data from AO3...", total=len(fic_ids))

        with ThreadPoolExecutor(max_workers=5) as executor:
            executor.map(load_fic, fic_ids)

    # Handle adding of each fic
    for count, fic in enumerate(external_fics):
        
        if 'Exception' in fic:
            add_row(fic, count)
            continue

        for value in fic:
            # If type of entry is list, store it as comma-seperated string (e.g. "foo, bar").
            if type(fic[value]) is list:
                fic[value] = ", ".join(fic[value])
            
        # Strip leading and trailing whitespace from fic summaries.
        fic['summary'] = fic['summary'].strip()
        
        if type(local_fics[count]['nchapters']) is type(None):
            add_row(fic, count)
        elif int(fic['nchapters']) > int(local_fics[count]['nchapters']):
            add_row(fic, count, styling=constants.UPDATED_STYLES)
        else:
            add_row(fic, count)

        # Convert every value into string so it can be stored in the database.
        for value in fic:
            fic[value] = str(fic[value])
        #fic = [str(fic[value]) for fic[value] in fic]

        # Update database
        database.update_fic(fic, fic['id'])

    # Print rich table
    print()
    console.print(table)

    table.title = "Fanfics (cached)"

    with open(constants.DATA_PATH + "table.pickle", 'wb') as file:
        pickle.dump(table, file, protocol=pickle.HIGHEST_PROTOCOL)


def add_url_multiple():
    message = click.edit(constants.MARKER + '\n')

    if message is not None:
        message_lines = message.split(constants.MARKER, 1)[1].rstrip('\n').lstrip('\n')
        urls_to_parse = message_lines.split('\n')

        for count, item in enumerate(urls_to_parse):
            entry_id = AO3.utils.workid_from_url(item)
            if entry_id == None:
                print(f"{item} is not a valid url.")
            elif str(entry_id) in fic_ids:
                print(item, "already in database.")
            else:
                database.add_fic(entry_id)
                print("Added " + urls_to_parse[count])

    construct_rich_table()


def add_url_single(entry):
    entry_id = AO3.utils.workid_from_url(entry)

    if entry_id == None:
        print(f"{entry} is not a valid url.")
    elif str(entry_id) in str(fic_ids):
        print(f"{entry_id} already in database.")
    else:
        database.add_fic(entry_id)
        print("Added", entry)

    construct_rich_table()


def delete_entry(entry):
    try:
        # Decrease entry by one because Python arrays start at 0.
        target_entry = local_fics[(entry - 1)]
        database.delete_fic(target_entry['id'])
    except IndexError:
        print("Number out of index range.")
        exit()
    print("Deleted entry number", str(entry))

    construct_rich_table()


def construct_rich_table(read_again=True):
    if read_again is True:
        # Read database again
        local_fics = database.get_all_fics()
    
    for count, fic in enumerate(local_fics):
        add_row(fic, count)

    # Print rich table
    print()
    console.print(table)


def print_cached_table():
    with open(constants.DATA_PATH + "table.pickle", 'rb') as file:
        table_cached = pickle.load(file)
        print()
        console.print(table_cached)


def add_row(fic, count, styling=""):
    # Copy the fic object so that changes made here are local, and not written to the database.
    fic = copy.copy(fic)

    # Offset index by 1 because Python arrays start at 0.
    index = str(count + 1)

    # Create AO3 link for fic
    fic_link = f"https://archiveofourown.org/works/{fic['id']}" 

    # If key 'Exception' in fic, display error information.
    if 'Exception' in fic:
        table.add_row(f"{index}.", f"[link={fic_link}]ERROR: {fic['Exception']}[/link]", style="red")
        return

    # Check if title or date_published field is empty. If it is, we assume that the fic has not yet been scraped.
    if fic['title'] is None or fic['date_published'] is None:
        table.add_row(f"{index}.", f"[link={fic_link}]FIC DATA NOT YET SCRAPED[/link]", style=styling)
        return

    # Converting expected_chapters from None to '?' makes the "Chapters" column look nicer.
    if fic['expected_chapters'] is None or fic['expected_chapters'] == "None":
        fic['expected_chapters'] = '?'
    
    # Turn upload date of fic into correct format 
    then = datetime.strptime(fic['date_updated'], constants.DATE_FORMAT)

    # Shorten date strings from YYYY-MM-DD XX:XX:XX to YYYY-MM-DD.
    fic['date_published'] = fic['date_published'][:-9]
    fic['date_edited'] = fic['date_edited'][:-9]
    fic['date_updated'] = fic['date_updated'][:-9]

    # Set $chapters to nchapters/expected_chapters (+difference)
    if styling == constants.UPDATED_STYLES:
        difference = int(fic['nchapters']) - int(local_fics[count]['nchapters'])
        fic['$chapters'] = f"{fic['nchapters']}/{fic['expected_chapters']} (+{difference})"
    else:
        fic['$chapters'] = f"{fic['nchapters']}/{fic['expected_chapters']}"

    # Get latest chapter title
    fic['$latest_chapter'] = fic['chapter_titles'][-1]

    # Trim every element to max_row_length
    for data in fic:
        if isinstance(fic[data], str):
            fic[data] = (fic[data][:constants.MAX_ROW_LENGTH].strip() + '...') if len(fic[data]) > constants.MAX_ROW_LENGTH else fic[data]

    new_args = []
    for count, zipped in enumerate(zip(args, constants.TABLE_TEMPLATE)):
        if constants.TABLE_TEMPLATE[count]['column'] == 'title':
            new_args.append(f"[link={fic_link}]{fic['title']}[/link]")
        else:
            new_args.append(fic[constants.TABLE_TEMPLATE[count]['column']])
    new_args.insert(0, f"{index}.")
    
    if (constants.NOW - then).days > constants.STALE_THRESHOLD:
        table.add_row(*new_args, style=constants.STALE_STYLES)
    else:
        table.add_row(*new_args, style=styling)


if __name__ == "__main__":
    main()



# This module contains important constants used in ao3scraper, and is imported by most files.

from os import path
from os import environ
from platformdirs import user_data_dir, user_config_dir
from datetime import datetime
from importlib import metadata
from yaml import load, Loader

# Custom modules
import ao3scraper.file_validator as file_validator

# Set important constants!
APP_NAME = "ao3scraper"
APP_AUTHOR = "EthanLeitch"
APP_VERSION = metadata.version(APP_NAME)
ALEMBIC_VERSION = 'ca2dfefaf0b6'

DATA_PATH = path.join(user_data_dir(APP_NAME, APP_AUTHOR)) + "/"
CONFIG_PATH = path.join(user_config_dir(APP_NAME, APP_AUTHOR)) + "/"

DATABASE_FILE_PATH = DATA_PATH + "fics.db"
CONFIG_FILE_PATH = CONFIG_PATH + "config.yaml"

# Set DATABASE_FILE_PATH environment variable for alembic
environ.setdefault("DATABASE_FILE_PATH", DATABASE_FILE_PATH)

CONFIG_TEMPLATE = """
# Formatting of table
max_row_length: 120
warnings: false
stale_threshold: 60
stale_styles: deep_sky_blue4
updated_styles: '#ffcc33 bold'

# Column attributes
table_template:
- column: title
  name: Title
  styles: magenta
- column: $chapters
  name: Chapters
  styles: green
- column: date_updated
  name: Last updated
  styles: cyan
- column: status
  name: Status
  styles: violet
"""

"""
Currently this list is only avaliable by fetching an instance of a Work's metadata (AO3.Work.metadata).
This is done in construct_list.py
'id' has been excluded from this list to prevent conflicts with the SQLAlchemy primary_key. 
"""
TABLE_COLUMNS = ['date_edited', 'date_published', 'date_updated', 'bookmarks', 'categories', 'nchapters', 'characters', 'complete', 'comments', 'expected_chapters', 'fandoms', 'hits', 'kudos', 'language', 'rating', 'relationships', 'restricted', 'status', 'summary', 'tags', 'title', 'warnings', 'words', 'collections', 'authors', 'series', 'chapter_titles']
CUSTOM_COLUMNS = ['$chapters', '$latest_chapter']

# Check that config.yaml and fics.db exist
file_validator.main()

# Load user's custom preferences from config.yaml
with open(CONFIG_FILE_PATH, 'r') as file:
    config_file = load(file, Loader=Loader)

# Load each preference as a constant variable
MAX_ROW_LENGTH = config_file['max_row_length']
WARNINGS = config_file['warnings']
STALE_THRESHOLD = config_file['stale_threshold']
STALE_STYLES = config_file['stale_styles']
UPDATED_STYLES = config_file['updated_styles']
TABLE_TEMPLATE = config_file['table_template']

# Check that all the columns listed in the config file are valid
for i in TABLE_TEMPLATE:
    if i['column'] not in TABLE_COLUMNS and i['column'] not in CUSTOM_COLUMNS:
        print(f"{i['column']} is not a valid column.")

# Other constants
MARKER = "# Enter one url on each line to add it to the database. This line will not be recorded."

DATE_FORMAT = "%Y-%m-%d %H:%M:%S"
NOW = datetime.now()


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['ao3scraper', 'ao3scraper.tests']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML==6.0',
 'Pygments==2.12.0',
 'SQLAlchemy>=1.4.41,<2.0.0',
 'alembic>=1.8.1,<2.0.0',
 'ao3-api>=2.3.0,<3.0.0',
 'beautifulsoup4==4.11.1',
 'certifi==2022.5.18.1',
 'chardet==4.0.0',
 'click==8.0.1',
 'commonmark==0.9.1',
 'configparser>=5.3.0,<6.0.0',
 'deepdiff[murmur]>=5.8.1,<6.0.0',
 'dictdiffer>=0.9.0,<0.10.0',
 'idna==2.10',
 'marshmallow-sqlalchemy>=0.28.1,<0.29.0',
 'pathlib>=1.0.1,<2.0.0',
 'platformdirs>=2.5.4,<3.0.0',
 'requests==2.25.1',
 'rich==12.4.1',
 'ruamel-yaml>=0.17.21,<0.18.0',
 'soupsieve==2.3.2.post1',
 'urllib3==1.26.9']

entry_points = \
{'console_scripts': ['ao3scraper = ao3scraper.cli:main']}

setup_kwargs = {
    'name': 'ao3scraper',
    'version': '1.0.2',
    'description': 'ao3scraper is a python webscraper that scrapes AO3 for fanfiction data, stores it in a database, and highlights entries when they are updated.',
    'long_description': "# ao3scraper\nA python webscraper that scrapes AO3 for fanfiction data, stores it in a database, and highlights entries when they are updated.\n\n![Fanfics Table](https://i.ibb.co/80r9vwR/Fanfic-Table.png)\n\n*Table with an updated entry highlighted.*\n\n## Installation\nYou can easily install the latest version from pip:\n\n    pip3 install ao3scraper\n\n### Development Installation\nCreate a python virtual environment with `python3 -m venv dev_venv` and activate it.\nThen, install required packages with:\n\n    poetry install\n\nThis will also install ao3scraper into the virtual environment.\n\n## Usage\n    Usage: ao3scraper [OPTIONS]\n\n    Options:\n    -s, --scrape          Launches scraping mode.\n    -c, --cache           Prints the last scraped table.\n    -l, --list            Lists all entries in the database.\n    -a, --add TEXT        Adds a single url to the database.\n    --add-urls            Opens a text file to add multiple urls to the database.\n    -d, --delete INTEGER  Deletes an entry from the database.\n    -v, --version         Display version of ao3scraper and other info.\n    --help                Show this message and exit.\n\n## Configuration\nao3scraper is ridiculously customisable, and most aspects of the program can be modified from here.\nTo find the configuration file location, run `python3 ao3scraper -v`.\n\nao3scraper uses [rich](https://rich.readthedocs.io/en/stable/style.html)'s styling. To disable any styling options, replace the styling value with 'none'.\n\nFics have many attributes that are not displayed by default. To add these columns, create a new option under table_template, like so:\n```yaml\ntable_template:\n- column: characters # The specified attribute\n  name: Characters :) # This is what the column will be labelled as\n  styles: none # Rich styling\n```\nA complete list of attributes can be found [on the wiki](https://github.com/EthanLeitch/ao3scraper/wiki/Fic-Attributes/).\n\n## Migrating the database\nIf you're updating from a legacy version of ao3scraper (before 1.0.0), move `fics.db` to the data location. \nThis can be found by running `python3 ao3scraper -v`.\nThe migration wizard will then prompt you to upgrade your database. \nIf you accept, a backup of the current `fics.db` will be created in `/backups`, and migration will proceed.\n\n## Contributing\nContributions are always appreciated. Submit a pull request with your suggested changes!\n\n## Acknowledgements\nao3scraper would not be possible without the existence of [ao3_api](https://github.com/ArmindoFlores/ao3_api/) and the work of its [contributors](https://github.com/ArmindoFlores/ao3_api/graphs/contributors).",
    'author': 'Ethan',
    'author_email': 'ethanjohnleitch@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/EthanLeitch/ao3scraper',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
